(* Sygnatura Expr (wyrazenia) i pasujaca do niej struktura, ktora *)
(* wykorzystuje typ Expr zdefiniowany w Types.sml.                *)

(* operatory infiksowe *)
infix 5 +++ ---;
infix 6 *** ///;

(* wyrazenia *)
signature Expr = sig
  
  (* wyrazenia real *)
  type expr;
  
  (* przypisania *)
  type assignment;
  
  structure Env : Env where type value = expr;
  
  (* funkcja *)
  type func = Env.env -> real;
    
  (* funkcja kolorowa *)
  type colorFunc = Env.env -> Image.color;

  (* proste wyrazenia *)
  val cst: real -> expr;
  val op +++ : (expr * expr) -> expr;
  val op --- : (expr * expr) -> expr;
  val op *** : (expr * expr) -> expr;
  val op /// : (expr * expr) -> expr;
  
  (* przypisz wartosc typu real *)
  val setc: Env.var -> real -> Env.env -> Env.env;
  (* pobierz wartosc typu real *)
  val getc: Env.env -> Env.var -> real;
  (* pobierz wartosc typu color *)
  val getcol: Env.env -> Env.var -> Image.color;
  (* uzyj przypisania *)
  val assign: assignment -> Env.env -> Env.env;
  (* zamiana wyrazenia (expr) na funkcje (func) *)
  val eval : expr -> func;
  (* zamiana wyrazenia (expr) na funkcje kolorowa (colorfunc) *)
  val evalcol : expr -> colorFunc;
    
  (* rozniczkowanie wyrazenia po podanej zmiennej; jesli jest niemozliwe, *)
  (* zostanie zwrocony wyjatek. Zmienna zazwyczaj powinna byc wolna w     *)
  (* podanym srodowisku, bo zmienne zdefiniowane beda rozwijane.          *)
  val D : Env.var -> expr -> Env.env -> expr;

  (* prowizoryczna definicja rozniczkowania ... *)
  axiom Forall (expr,env,var,x) => (Env.undefined env var) implies let
    fun lim0 f = f(0.001) (* "granica przy h->0" *)
    fun ev expr x = eval expr (Env.set var (cst (x)) env)
    fun diff h = (ev expr (x+h) - ev expr x) / h
    in Env.undefined env var implies
      ev (D var expr env) x = lim0 diff
    end;

  (* upraszczanie: ewaluacja stalych, podstawianie zmiennych zdefiniowanych *)
  (* w srodowisku, usuwanie niepotrzebnych operacji typu "...+0" itp.       *)
  val simplify: Env.env -> expr -> expr;
    
  (* upraszczanie ma nie zmieniac wartosci ... *)
  axiom Forall (expr,env1,env2) => let
    fun extension env1 env2 = 
      Forall(var,x) =>
        (Env.undefined env1 var) orelse
	(Env.defined env2 var implies Env.eq(Env.get env1 var, Env.get env2 var))
    in extension env1 env2 implies (
      (eval expr env2 = eval (simplify env1 expr) env2)
      andalso
      (evalcol expr env2 = evalcol (simplify env1 expr) env2)
      )
    end;

  (* postaraj sie wypisac wyrazenie *)
  val tostring: expr -> string;
  end;

(***************************************************************************)

functor Expr(E: Env where type var = Var and type value = Expr) : Expr = struct

  structure Env = E;
  
  type expr = Expr;
  type assignment = Assignment;
  type func = Env.env -> real;
  type colorFunc = Env.env -> Image.color;

  fun cst x = EC x;
  fun op +++ (x,y) = x++y;
  fun op --- (x,y) = x--y;
  fun op *** (x,y) = x**y;
  fun op /// (x,y) = x//y;
  
  fun setc v x e = Env.set v (EC x) e;
  
  fun ev (a++b) e  = ev a e + ev b e
    | ev (a--b) e  = ev a e - ev b e
    | ev (a**b) e  = ev a e * ev b e
    | ev (a//b) e  = ev a e / ev b e
    | ev (EC c) e  = c
    | ev (EV v) e  = getc e v
    | ev (EFSin  a) e = Math.sin  (ev a e)
    | ev (EFCos  a) e = Math.cos  (ev a e)
    | ev (EFSqrt a) e = Math.sqrt (ev a e)
    | ev (EFSqr  a) e = square    (ev a e)
    | ev (EFExp  a) e = Math.exp  (ev a e)
    | ev (EFLn   a) e = Math.ln   (ev a e)
    | ev (EFAtan a) e = Math.atan (ev a e)
    | ev (EFAny(f,a)) e = f(ev a e)
    | ev (EIf (c,t,f)) e = ev (if ev c e>0.0 then t else f) e
    | ev (ELocal(a,asg)) e = ev a (assign asg e)
    | ev (ERed(a)) e = #1 (evc a e)
    | ev (EGreen(a)) e = #2 (evc a e)
    | ev (EBlue(a)) e = #3 (evc a e)
    | ev (ENum(a)) c = ev a c
    | ev c e = let
      val (r,g,b) = evc c e
      in 0.299*r+0.587*g+0.114*b
      end
  and evc (a++b) e = add3 (evc a e, evc b e)
    | evc (a--b) e = sub3 (evc a e, evc b e)
    | evc (a**b) e = mul3 (ev a e, evc b e)
    | evc (EIf(c,t,f)) e = evc (if ev c e>0.0 then t else f) e
    | evc (ELocal(a,asg)) e = evc a (assign asg e)
    | evc (EV v) e  = getcol e v
    | evc (ERGB(r,g,b)) e = (ev r e, ev g e, ev b e)
    | evc (EImage(img, ex, ey)) e = let
      val sx = Image.xsize img
      val sy = Image.ysize img
      val x  = ev ex e
      val y  = ev ey e
      val nx = (Real.trunc((Real.fromInt sx)*(x+0.999999)/2.0)) mod sx
      val ny = (Real.trunc((Real.fromInt sy)*(y+0.999999)/2.0)) mod sy
      in Image.getpixel(img,nx,ny)
      end
    | evc (ENum(a)) c = evc a c
    | evc a e = (ev a e, ev a e, ev a e)
  and eval e = ev e
  and evalcol e = evc e
  and getc e v = ev (Env.get e v) e
  and getcol e v = evc (Env.get e v) e
  and assign (AFn (x,a)) e = Env.set x a e
    | assign (ACst(x,a)) e = setc x (eval a e) e
    | assign (ANul(x)) e = Env.undef x e;

  fun D v1 (a++b)    e = D v1 a e ++ D v1 b e
    | D v1 (a**b)    e = (D v1 a e ** b) ++ (a ** D v1 b e)
    | D v1 (a--b)    e = D v1 a e -- D v1 b e
    | D v1 (a//b)    e = 
      (D v1 a e//b) -- (a ** D v1 b e // (b ** b))
    | D v1 (EC c)    e = EC 0.0
    | D v1 (EFSin a) e = D v1 a e ** EFCos a
    | D v1 (EFCos a) e = D v1 a e ** (EC 0.0 -- EFSin a)
    | D v1 (EFExp a) e = D v1 a e ** EFExp a
    | D v1 (EFLn  a) e = D v1 a e ** (EC 1.0 // a)
    | D v1 (EFSqrt a) e = D v1 a e ** (EC 0.5 // EFSqrt a)
    | D v1 (EFSqr a) e = D v1 a e ** EC 2.0 ** a
    | D v1 (EFAtan a) e = D v1 a e // (EC 1.0 ++ EFSqr a)
    | D v1 (EV v2)   e = if Env.defined e v2
      then (D v1 (Env.get e v2) e)
      else (if v1=v2 then EC 1.0 else EC 0.0)
    | D v1 (EIf (c,t,f)) e = EIf(c,D v1 t e,D v1 f e)
    | D v1 (ELocal(a,asg as AFn(x,a1))) e = ELocal(D v1 a (assign asg e), asg)
    | D v1 (ELocal(a,asg as ANul(x))) e = ELocal(D v1 a (assign asg e), asg)
    | D v1 (ELocal(a,asg as ACst(x,a1))) e = 
      ELocal(D v1 a (Env.undef x e), ACst(x,a1)) ++ (D v1 a1 e) ** (D x a (Env.undef x e))
    | D v1 (ENum(a)) e =
      (EC 1000.0) ** ENum(ELocal(a, ACst(v1, EV v1 ++ EC 0.0001)) -- a)
    | D v1 _         e = raise Fail "Differentiation not implemented"

  (* jeden krok upraszczania *)
  
  fun simpsin ((EC 0.0) ++ b)    = b
    | simpsin (a ++ (EC 0.0))    = a
    | simpsin (a ** (EC 0.0))    = EC 0.0
    | simpsin ((EC 0.0) ** a)    = EC 0.0
    | simpsin (a ** (EC 1.0))    = a
    | simpsin ((EC 1.0) ** a)    = a
    | simpsin (a -- (EC 0.0))    = a
    | simpsin (a // (EC 1.0))    = a
    | simpsin ((EC 0.0) // a)    = EC 0.0
    | simpsin ((EC a) ++ (EC b)) = EC (a+b)
    | simpsin ((EC a) -- (EC b)) = EC (a-b)
    | simpsin ((EC a) ** (EC b)) = EC (a*b)
    | simpsin ((EC a) // (EC b)) = EC (a/b)
    | simpsin (EFSin  (EC a))    = EC (Math.sin  a)
    | simpsin (EFCos  (EC a))    = EC (Math.cos  a)
    | simpsin (EFExp  (EC a))    = EC (Math.exp  a)
    | simpsin (EFLn   (EC a))    = EC (Math.ln   a)
    | simpsin (EFAtan (EC a))    = EC (Math.atan a)
    | simpsin (EFSqrt (EC a))    = EC (Math.sqrt a)
    | simpsin (EFSqr  (EC a))    = EC (a*a)
    | simpsin (EFAny  (f,EC a))  = EC (f a)
    | simpsin expr               = expr;

  fun
    simplify e (a ++ b)         = simpsin (simplify e a ++ simplify e b)
  | simplify e (a -- b)         = simpsin (simplify e a -- simplify e b)
  | simplify e (a ** b)         = simpsin (simplify e a ** simplify e b)
  | simplify e (a // b)         = simpsin (simplify e a // simplify e b)
  | simplify e (EFSin  a)       = simpsin (EFSin  (simplify e a))
  | simplify e (EFCos  a)       = simpsin (EFCos  (simplify e a))
  | simplify e (EFExp  a)       = simpsin (EFExp  (simplify e a))
  | simplify e (EFLn   a)       = simpsin (EFLn   (simplify e a))
  | simplify e (EFAtan a)       = simpsin (EFAtan (simplify e a))
  | simplify e (EFSqrt a)       = simpsin (EFSqrt (simplify e a))
  | simplify e (EFSqr a)        = simpsin (EFSqr  (simplify e a))
  | simplify e (EFAny(f,a))     = simpsin (EFAny  (f,simplify e a))
  | simplify e (ELocal(a,asg as AFn(_,_)))  = simplify (assign asg e) a
  | simplify e (ELocal(a,asg as ACst(x,a1)))  =
    ELocal(simplify (Env.undef x e) a, ACst(x, simplify e a1))
  | simplify e (ELocal(a,asg as ANul(x)))  =
    ELocal(simplify (Env.undef x e) a, ANul(x))
  | simplify e (ERGB(r,g,b))    = ERGB(simplify e r, simplify e g, simplify e b)
  | simplify e (EIf(c,t,f))     = EIf(simplify e c, simplify e t, simplify e f)
  | simplify e (EImage(img,x,y))= EImage(img,simplify e x,simplify e y)
  | simplify e (EV v) = 
    if Env.defined e v then simplify e (Env.get e v) else EV v
  | simplify e (ENum a)         = ENum a
  | simplify e expr = expr;

  fun ts (EC c) = rts c
    | ts (a ++ b)  = "("^(ts a)^"+"^(ts b)^")"
    | ts (a -- b)  = "("^(ts a)^"-"^(ts b)^")"
    | ts (a ** b)  = "("^(ts a)^"*"^(ts b)^")"
    | ts (a // b)  = "("^(ts a)^"/"^(ts b)^")"
    | ts (EFSin  a) = "sin("^(ts a)^")"
    | ts (EFCos  a) = "cos("^(ts a)^")"
    | ts (EFExp  a) = "exp("^(ts a)^")"
    | ts (EFLn   a) = "ln("^(ts a)^")"
    | ts (EFAtan a) = "atan("^(ts a)^")"
    | ts (EFSqrt a) = "sqrt("^(ts a)^")"
    | ts (EFSqr a)  = "sqr("^(ts a)^")"
    | ts (EFAny(f,a))="fxxx("^(ts a)^")"
    | ts (ELocal(a,AFn(x,ex))) = "("^(ts a)^")["^varname x^"="^(ts ex)^"]"
    | ts (ELocal(a,ACst(x,ex))) = "("^(ts a)^")["^varname x^":="^(ts ex)^"]"
    | ts (ELocal(a,ANul(x))) = "("^(ts a)^")["^varname x^"=?]"
    | ts (ERGB(r,g,b)) = "("^(ts r)^","^(ts g)^","^(ts b)^")"
    | ts (EIf(c,t,f)) = "?("^(ts c)^","^(ts t)^","^(ts f)^")"
    | ts (EImage(img,x,y)) = "<image>("^(ts x)^","^(ts y)^")"
    | ts (EV v) = varname v
    | ts (ENum a) = "num("^(ts a)^")"
    | ts _ = "<???>";
    
  fun tostring x = ts x;
  end;
