(* interpreter *)

structure Main :> Main = struct

  (* parsowanie instrukcji i wyrazen *)
  fun parseInstr str = let
    val lex = Lexing.createLexerString str
    val res:Instr = Grammar.InstrEOF Token lex
    in  res
    end;

  fun parseExpr str = let
    val lex = Lexing.createLexerString str
    val res:Expr = Grammar.ExprEOF Token lex
    in  res
    end;

  structure Env = Env(type var = Var; type value = Expr);
  
  structure Expr = Expr(Env);
  
  open Env;
  open Expr;
  open Image;
  
  type Result = unit;

  (* stan *)
  (*------*)

  (* stan zapisujemy jako n-tki obiektow roznych typow *)
  datatype State = STATE of (
    env * (int*int) * Image.image * R3.transform * 
    Scene.scene * (State->Result) *unit*unit);
    
  (* To so funkcje do brania/ustawiania jednej z czesci n-tki *)
  fun getA (STATE(a,b,c,d,e,f,g,h)) = a;
  fun getB (STATE(a,b,c,d,e,f,g,h)) = b;
  fun getC (STATE(a,b,c,d,e,f,g,h)) = c;
  fun getD (STATE(a,b,c,d,e,f,g,h)) = d;
  fun getE (STATE(a,b,c,d,e,f,g,h)) = e;
  fun getF (STATE(a,b,c,d,e,f,g,h)) = f;
  fun getG (STATE(a,b,c,d,e,f,g,h)) = g;
  fun getH (STATE(a,b,c,d,e,f,g,h)) = h;

  fun setA (STATE(a,b,c,d,e,f,g,h)) x = (STATE(x,b,c,d,e,f,g,h));
  fun setB (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,x,c,d,e,f,g,h));
  fun setC (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,b,x,d,e,f,g,h));
  fun setD (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,b,c,x,e,f,g,h));
  fun setE (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,b,c,d,x,f,g,h));
  fun setF (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,b,c,d,e,x,g,h));
  fun setG (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,b,c,d,e,f,x,h));
  fun setH (STATE(a,b,c,d,e,f,g,h)) x = (STATE(a,b,c,d,e,f,g,x));
  
  val getEnv = getA; val setEnv = setA;
  val getSize = getB; val setSize = setB;
  val getBG = getC; val setBG = setC;
  val getRot = getD; val setRot = setD;
  val getScene = getE; val setScene = setE;
  val getQuitCont = getF; val setQuitCont = setF;

  (* tworzenie SurfaceParams na podstawie stanu *)
  
  functor SurfaceParams (val s: State) : SurfaceParams = struct
    structure Expr = Expr;
    type expr  = expr;
    type env   = env;
    type expr3 = expr R3.up;
    type func  = Expr.func;
    type colorFunc = Expr.colorFunc;
    
    val env = getEnv s;

    val sizes as (CX,CY) = (xsize (getBG s), ysize (getBG s));
    
    val size = if CX<CY then CX else CY;
    val scs  = (getc env VScale) * itr size;

    fun scale(x,y,z) =
      (0.5*(scs*x + itr CX),
       0.5*(scs*y + itr CY),
       scs*z);

    val rotate = getRot s;
    
    end;
  
  (* srodowisko z ustawionym u i v *)
  fun uvenv (u,v) env = setc VU u (setc VV v env);

  (* kontynuacja *)
  type Cont = State -> Result;
  
  (* narysuj tlo, wzor podany w srodowisku *)
  fun dodrawb env (sx,sy) = let
    fun xfor 0 f = f(0) | xfor x f = (f(x-1); xfor (x-1) f);

    fun xyfor x y f = xfor y (fn yy => xfor x (fn xx => f (xx,yy)));

    val img = Image.newimage(sx,sy);

    fun itrx x = ((itr x / itr sx) * 2.0) - 1.0;
    fun itry y = ((itr y / itr sy) * 2.0) - 1.0;

    in (xyfor sx sy (fn (x,y) =>
      Image.putpixel(img, x, y, getcol (uvenv ((itrx x, itry y)) env) VCBkg)
      ); img)

    end;

  (* przerotuj 'orig' o kat 'x' rotacja 'rot' *)
  fun rotate orig rot x = let

    val C = Math.cos (x*Math.pi/180.0)
    val S = Math.sin (x*Math.pi/180.0)
    
    fun ro RotXY (x,y,z) = (x*C-y*S, y*C+x*S, z)
      | ro RotYZ (x,y,z) = (x, y*C-z*S, z*C+y*S)
      | ro RotZX (x,y,z) = (x*C+z*S, y, z*C-x*S)
    in (ro rot) o orig

    end;

  (* wczytaj linie z is *)
  fun fgetline is =
    let val line = TextIO.inputLine is
    in String.substring (line, 0, size line-1)
    end;

  (* wczytaj linie ze stdin *)
  fun getline () = fgetline TextIO.stdIn

  (* znajdywanie w strumieniu is linii z podana trescia *)
  fun findLine is line = if TextIO.endOfStream is then false else let
    val s = fgetline is
    in if s = line then true else findLine is line
    end;

  (* wypisywanie helpu, az do nastepnego tematu *)
  fun displayHelp is = if TextIO.endOfStream is then () else let
    val s = TextIO.inputLine is
    in if s <> "" andalso String.sub(s,0) = #"@" then () else (
      print s;
      displayHelp is
      )
    end;

  (* wypisz blad *)
  fun printEx e = (
    print (case e of 
          Undefined x => "Undefined variable: "^(varname x)
        | Interrupt => "Interrupted"
        | Parsing.ParseError _ => "Parse error: "
        | Fail x => "Fail: "^x
        | Div => "Division by zero"
        | Overflow => "Overflow" 
	| Domain => "Domain error"
        | e => "Error: " ^ (exnMessage e)
	);
    print ".\n"
    );

  (* operator stalopunktowy *)  
  fun Fix F = F (fn s => Fix F s);

  (* tryb interaktywny, ze stanu s*)
  fun interactive s = let
    val _ = print "Enter `quit' to quit.\n";
    fun F cont s = let
      val _ = print "surface> "
      val line = getline ()
      val instr =
        parseInstr line handle e => (printEx e; INull)
      in doI instr cont s
      end
    in Fix F s
    end

  (* semantyka instrukcji i *)
  and doI (i:Instr) (c:Cont) (s:State) : Result = let
    val env = getEnv s
    fun evalc e = eval e env

    (* funkcja S zwraca kontynuacje i nastepny stan. Nie aplikuje kontynuacji *)
    (* do stanu --- w ten sposob wyjatki wystepujace wewnatrz kontynuacji nie *)
    (* beda lapane                                                            *)
    fun
      DoI (INull) = (c,s) |
      DoI (ISeq(i1,i2)) = (doI i1 (doI i2 c), s) |  
      DoI (IQuit) = (getQuitCont s,s) |
      DoI (IInteractive) = let
        val qc = getQuitCont s
	fun qc2 s2 = c (setQuitCont s2 qc)
	in (interactive, setQuitCont s qc2)
	end |
      DoI (IIf(g,t,f)) = (doI (if evalc g>0.0 then t else f) c, s) |
      DoI (IWhile(g,t)) = let
        fun F c1 s = if eval g (getEnv s)>0.0 then doI t c1 s else c s
        in (Fix F, s)
        end |
(*    DoI (IWhile(g,t)) = (doI (IIf(g,ISeq(t,IWhile(g,t)),INull)) c, s) | *)
      DoI (IInclude(file)) = let
        val is = TextIO.openIn file
	val str = TextIO.inputAll is
	val _ = TextIO.closeIn is
	val instr = parseInstr str
	in (doI instr c, s)
        end |
      DoI (IHelp(topic)) = let
        val is = TextIO.openIn "surface.hlp"
	val _ = if findLine is ("@"^topic)
	  then (
	    print "\n";
	    displayHelp is
	    )
	  else print "Topic not found\n"
	val _ = TextIO.closeIn is
	in (c, s)
        end |
      DoI (IValue(e)) = let
        val vl = rts(evalc e)
	val _ = print (">> " ^ vl ^ "\n")
	in (c, s)
	end |
      DoI (IDiff(v,e)) = let
(*      val _ = print ("*> " ^ (tostring (D v e env)) ^ "\n") *)
        val _ = print ("*> " ^ (tostring (simplify env (D v e env))) ^ "\n")
	in (c,s)
        end |
      DoI (IInfo) = let
        fun its r = Int.toString r
	val (x,y) = getSize s
	val (sx,sy) = (its x, its y)
	val img = getBG s
	val (tx,ty) = (its (Image.xsize img), its (Image.ysize img))
	val ss = (its (length (getScene s)))
	val _ = print ("Size " ^ sx ^ "x" ^ sy ^ " ("^tx^"x"^ty^"), scene: "^ss^" pixels\n")
	in (c, s)
	end |
      
      DoI (ISize(ex,ey)) = (c, setSize s (Real.ceil (evalc ex), Real.ceil (evalc ey))) |
      DoI (IDrawBG) =
        (print "Painting...\n"; (c, setBG s (dodrawb env (getSize s)))) |
      DoI (ILoad(fname)) = (c, setBG s (Image.loadbmp(fname))) |
      
      DoI (IClear) = (c, setScene s nil) |
      
      DoI (ISurf) = let
        val _ = print "Drawing";
	structure Surface = Surface(SurfaceParams(val s = s));
	val scn = Surface.surface ();
        val _ = print "\n";
	in (c, setScene s (scn @ (getScene s)))
	end |
  
      DoI (ICurve) = let
        val _ = print "Drawing";
	structure Surface = Surface(SurfaceParams(val s = s));
	val scn = Surface.curve ();
        val _ = print "\n";
	in (c, setScene s (scn @ (getScene s)))
	end |
  
      DoI (IPoint(x,y,z)) = let
        val _ = print "Drawing...\n";
	structure Surface = Surface(SurfaceParams(val s = s));
	val scn = Surface.point (evalc x, evalc y, evalc z);
	in (c, setScene s (scn @ (getScene s)))
	end |
  
      DoI (IGeo((x0,y0),(dx,dy),len)) = let
        val _ = print "Drawing...\n";
	structure Surface = Surface(SurfaceParams(val s = s));
   	open Surface;
	val scn = geodesic ((evalc x0,evalc y0),(evalc dx,evalc dy),evalc len)
	in (c, setScene s (scn @ (getScene s)))
	end |

      DoI (IMerge) = (print "Merging...\n";
        (c, setScene (setBG s (Scene.drawscene (getScene s) (getBG s))) nil)) |
	
      DoI (ISave(file)) = let
        val _ = print "Merging...";
        val merge = Scene.drawscene (getScene s) (getBG s);
        val _ = print "\^H\^H\^H and saving...\n"; 
        val _ = Image.savebmp (file, merge);
	in (c, s)
	end |
      
      DoI (IView(file)) = (print "Viewing...\n"; 
        Process.system ("viewer "^file);
	(c,s)) |
      
      DoI (ISet(a)) = (c, setEnv s (assign a env)) |
      
      DoI (IRot0) = (c, setRot s R3.Id) |

      DoI (IRot(rt, a)) = (c, setRot s 
        (rotate (getRot s) rt (evalc a)));

      (*| DoI _ = (print "Unimplemented.\n"; c s);*)

    (* najpierw lapiemy wyjatki... *)
    val (c,s) = DoI i handle e => (
      printEx e;
      (c,s)
      );
    (* ... potem ogonowo aplikujemy --- w ten sposob handlery sie          *)
    (* nie kumuluja, co powinno pozwolic na lepsze dzialanie odsmiecania   *)
    in c s
    end;
  
  (* kontynuacja, ktora nic nie robi *)
  fun doNothing s = ();
  
  local
  
(*val fx = EV VU;
  val fy = EV VV;
  val fz = EC 0.0;
  val fz = parseExpr "(u*u+v*v)*0.8-0.5";
  val fz = parseExpr "0-1+2*exp((0-10)*(u*u+v*v))";
  val fz = parseExpr "((sin(4*u))+(cos(5*v)))/2";*)

  val env = empty
  val env = set VX (parseExpr "sin(pi*u/2)") env
  val env = set VY (parseExpr "(cos(pi*u/2))*sin(pi*v)") env
  val env = set VZ (parseExpr "(cos(pi*u/2))*cos(pi*v)") env
  val env = setc VScale 0.8 env
  val env = set VCBkg (parseExpr "(0,0,(u*u+v*v)/2)") env
  val env = setc VThick 2.0 env
  val env = setc VWrap 1.0 env
  val env = set VCSrf (parseExpr "(fcos(v), fcos(u), 0)") env
  val env = set VCGeo (parseExpr "(1,1,t)") env
  val env = set VCCrv (parseExpr "(fbow(t),fbow(t+0.3333),fbow(t+0.6666))") env
  val env = set VCPnt (parseExpr "(1,1,1)") env
  val env = setc VBig 1.0 env

  val size = (128,128)
  (* rot0; rotxy 45; rotyz 60 *)


  val rot = rotate (rotate R3.Id RotXY 45.0) RotYZ 60.0

  in
  val initstate = STATE(env, size, dodrawb env size, rot, nil, doNothing,(),())
  end;
  
  end;
