(* Dodawanie powierzchni i geodezyjnych do sceny. *)

(* Sygnatura SurfaceParams okresla parametry powierzchni, takie, jak        *)
(* srodowisko (w tym xyz), rotacja, skalowanie. Funktor Surface przeksztalca*)
(* parametry SurfaceParams na strukture o sygnaturze Surface, w ktorej sa   *)
(* funkcje rysujace powierzchnie i geodezyjne.                              *)

(* parametry *)
signature SurfaceParams = sig

  (* typ wyrazen *)
  structure Expr: Expr where type Env.var = Var.Var;

  type env = Expr.Env.env;
  type expr = Expr.expr;
  type func = Expr.func;
  type colorFunc = Expr.colorFunc;
  type expr3 = expr R3.up;

  (* srodowisko *)
  val env: env;
  
  (* clipping (obcinanie do prostokata) *)
  val CX: int;
  val CY: int;
  
  (* rotacja *)
  val rotate: R3.transform;
  
  (* skala *)
  val scale: R3.transform;
  
  end;

(* rysowanie *)
signature Surface = sig
  structure Params: SurfaceParams;
  
  (* narysuj powierzchnie *)
  val surface: unit -> Scene.scene;

  (* narysuj geodezyjna *)
  val geodesic: (R2.Vector.t*R2.Vector.t*real) -> Scene.scene;

  (* narysuj punkt  *)
  val point: R3.Vector.t -> Scene.scene;

  (* narysuj krzywa *)
  val curve: unit -> Scene.scene;
  end;

(***************************************************************************)

(* funktor Surface *)
functor Surface (X:SurfaceParams) : Surface = struct

  structure Params = X;
  
  open Params;
  open Scene;
  open Expr;
  open Env;
  open Var;
  
  (* bedziemy rozniczkowac i upraszczac po u i v, zatem... *)
  val env = undef VU env;
  val env = undef VV env;
  
  (* wspolrzedne ... *)
  val (ex,ey,ez) = (get env VX, get env VY, get env VZ);

  (* obliczanie funkcji w punkcie *)
  fun uvenv (u,v) env = setc VU u (setc VV v env);
  fun eval f (u,v) = Expr.eval f (uvenv (u,v) env);

  (* cache: od razu upraszczamy i obliczamy pochodne, zeby nie musiec *)
  (* robic tego pozniej za kazdym razem                               *)
  
  (* x,y,z *)
  val getX  = eval ex;
  val getY  = eval ey;
  val getZ  = eval ez;
  
  (* pierwsze pochodne *)
  val getUX = eval (simplify env (D VU ex env))
  val getUY = eval (simplify env (D VU ey env))
  val getUZ = eval (simplify env (D VU ez env))
  fun getU p = (getUX p, getUY p, getUZ p)

  val getVX = eval (simplify env (D VV ex env))
  val getVY = eval (simplify env (D VV ey env))
  val getVZ = eval (simplify env (D VV ez env))
  fun getV p = (getVX p, getVY p, getVZ p)
  
  (* drugie pochodne *)
  val getUUX = eval (simplify env (D VU (D VU ex env) env))
  val getUUY = eval (simplify env (D VU (D VU ey env) env))
  val getUUZ = eval (simplify env (D VU (D VU ez env) env))
  val getUVX = eval (simplify env (D VU (D VV ex env) env))
  val getUVY = eval (simplify env (D VU (D VV ey env) env))
  val getUVZ = eval (simplify env (D VU (D VV ez env) env))
  val getVVX = eval (simplify env (D VV (D VV ex env) env))
  val getVVY = eval (simplify env (D VV (D VV ey env) env))
  val getVVZ = eval (simplify env (D VV (D VV ez env) env))
  fun getUU p = (getUUX p, getUUY p, getUUZ p)
  fun getUV p = (getUVX p, getUVY p, getUVZ p)
  fun getVV p = (getVVX p, getVVY p, getVVZ p)

  (* typ przechowujacy punkt na powierzchni -- parametry *)
  (* (u,v) razem ze wspolrzednymi (x,y,z)                *)
  
  type surfacepoint = (real * real) * (real * real * real);
  
  fun surfpoint (u,v) = ((u,v),(getX(u,v),getY(u,v),getZ(u,v)))

  (* operacje na surfacepointach: *)
  
  (* podstawowe *)
  fun getsU ((u, v), (x, y, z)) = u;
  fun getsV ((u, v), (x, y, z)) = v;
  fun getsX ((u, v), (x, y, z)) = x;
  fun getsY ((u, v), (x, y, z)) = y;
  fun getsZ ((u, v), (x, y, z)) = z;
  fun getsXYZ (uv, xyz) = xyz;
  (* wektor ortogonalny w sp *)
  fun getorthogonal (p, xyz) = crossprod (getU p) (getV p);
  (* wektor ortonormalny w sp *)
  fun getorthonormal sp = R3.normalize (getorthogonal sp);

  (* wzbogacone srodowisko w surfacepoincie *)
  fun envn (sp as (uv,xyz)) = let
    (* macierz potrzebna do liczenia krzywizny *)
    fun M() = let
      val Du = getU uv 
      val Dv = getV uv
      val n  = getorthogonal sp
      val guu = R3.dotprod Du Du
      val guv = R3.dotprod Du Dv
      val gvv = R3.dotprod Dv Dv
      val G =  ((guu,guv),(guv,gvv)) (* metryka Riemanna *)
      val buu = R3.dotprod n (getUU uv)
      val buv = R3.dotprod n (getUV uv)
      val bvv = R3.dotprod n (getVV uv)
      val B =  ((buu,buv),(buv,bvv)) (* druga forma podstawowa *)
      in R2.matrix((R2.transform (R2.inverse G)) o (R2.transform B))
      end;
    val M = Susp.delay M;
    val env = env;
    val env = setc VU (getsU sp) env;
    val env = setc VV (getsV sp) env;
    val env = setc VX (getsX sp) env;
    val env = setc VY (getsY sp) env;
    val env = setc VZ (getsZ sp) env;
    val env = setLazy VS (fn () => cst(R3.norm (getorthogonal sp))) env;
    val env = setLazy VK (fn () => cst(R2.det(Susp.force M))) env;
    val env = setLazy VH (fn () => cst(R2.tr (Susp.force M))) env;
    in env
    end;

  fun center uva uvb = mul2(0.5, add2(uva, uvb))

  (* Funkcja addpoint dodaje punkt do sceny. Jednym z jej parametrow *)
  (* jest dim, ktory okresla, czy ten punkt nalezy do powierzchni    *)
  (* (2), krzywej (1) czy punktu (0). Dzialanie tego parametru jest  *)
  (* takie, ze np. dla krzywych do wspolrzednej z dodajemy 2, zeby   *)
  (* byla rysowana na powierzchni, a nie pod nia. Dla punktow doda-  *)
  (* jemy 4. (Za mala liczba powoduje, ze czasem mimo to rysujemy    *)
  (* pod powierzchnia, za duza --- ze czasem rysujemy nad jej inna   *)
  (* czescia; 2 powinno byc dobra liczba). Punkty nalezace do krzy-  *)
  (* wych i samodzielne sa rysowane jako duze, jesli big>0.          *)

  val big = getc env VBig > 0.0;
  fun addpoint1 (x,y,z,s) (scn:scene) = if big then
        (x,y,z,s)::(x+1,y,z,s)::(x-1,y,z,s)::(x,y+1,z,s)::(x,y-1,z,s)::scn
      else
        (x,y,z,s)::scn

  fun addpoint 2 (x,y,z,s) (scn:scene) = (x,y,z,s)::scn
    | addpoint 1 (x,y,z,s) (scn:scene) = addpoint1 (x,y,z+2.0,s) scn
    | addpoint 0 (x,y,z,s) (scn:scene) = addpoint1 (x,y,z+4.0,s) scn
    | addpoint dim  _  scn = scn

  (* powierzchnia *)
  
  fun surface () = let
  
    (* rysowanie konkretnego surfacepointa w obliczonym *)
    (* wczesniej punkcie (xx:int,yy:int)                *)
    fun scpix (xx,yy) sp = let
      val (vx,vy,vz) = rotate(getorthonormal sp)
      val ((u,v),xyz) = sp
      val (x,y,z) = xyz
      val zz = #3(scale(rotate(xyz)))
      val envn = envn sp
      val thick = getc envn VThick
      val col   = getcol envn VCSrf
      val trans =
        if thick=0.0 then 1.0 else
          if vz>(~0.01) andalso vz<0.01 then 0.0 else Math.exp(~(abs(thick/vz)))
      in (xx, yy, zz, (mul3(1.0-trans, col), trans)) : scenepixel
      end;

    (* Algorytm rysowania powierzchni musi dzialac w    *)
    (* ten sposob, ze jesli linia wzroku o grubosci 1   *)
    (* piksela przecina powierzchnie w N miejscach, to  *)
    (* w scenie w tym miejscu powinno sie pojawic N     *)
    (* scenepikseli (wymaganie to nie musi byc spelnio- *)
    (* ne perfekcyjnie). Dziala on tak:                 *)

    (* Wiekszosc operacji wykonuje funkcja triangle,    *)
    (* ktora do podanej sceny dorysowuje trojkat o      *)
    (* podanych wierzcholkach. Glowna funkcja dzieli    *)
    (* kwadrat na dwa trojkaty i wywoluje dla nich      *)
    (* triangle.                                        *)

    (* Funkcja triangle sprawdza, czy podany trojkat    *)
    (* jest taki maly, ze moze byc w nim co najwyzej 1  *)
    (* srodek piksela. Jesli nie, to rekurencyjnie      *)
    (* dzieli trojkat na trojkaty o mniejszej srednicy, *)
    (* ktorych jest 4. Jesli sie miesci, to sprawdza,   *)
    (* czy ten srodek jest we wnetrzu trojkata --- jesli*)
    (* tak, to go rysuje, a jesli nie, to nie.          *)
    
    (* To nie dziala perfekcyjnie, ale bardzo dobrze.   *)

    fun triangle step scn
      (sp1 as (uv1,xyz1)) (sp2 as (uv2,xyz2)) (sp3 as (uv3,xyz3)) = let
      val (x1,y1,_) = scale(rotate(xyz1))
      val (x2,y2,_) = scale(rotate(xyz2))
      val (x3,y3,_) = scale(rotate(xyz3))
      (* ograniczenia kwadratu *)
      val xm = Real.ceil (min3 x1 x2 x3) val xM = Real.floor (max3 x1 x2 x3)
      val ym = Real.ceil (min3 y1 y2 y3) val yM = Real.floor (max3 y1 y2 y3)
      in if xm>=xM andalso ym>=yM andalso step>3 (* maly trojkat *)
        then (* sprawdz clipping i czy punkt jest w srodku *)
	  (if (xm=xM) andalso (ym=yM) andalso 
          (xm>0) andalso (xm<CX) andalso (ym>0) andalso (ym<CY) andalso let
          val xx = Real.fromInt xm
          val yx = Real.fromInt ym
          in intriangle (x1,y1,x2,y2,x3,y3,xx,yx)
          end
          then (scpix (xm,ym) sp1) :: scn
	  else scn)
        else let (* podziel trojkat *)
          val spx1 = surfpoint(center uv2 uv3)
          val spx2 = surfpoint(center uv3 uv1)
          val spx3 = surfpoint(center uv1 uv2)
          in triangle (step+1) (triangle (step+1) 
	    (triangle (step+1) (triangle (step+1) scn 
            spx1 spx2 spx3) sp1 spx2 spx3) spx1 sp2 spx3) spx1 spx2 sp3
          end
      end

    (* kwadrat to 2 trojkaty *)
    fun square x1 y1 x2 y2 scn = let
      val ea = surfpoint(x1,y1)
      val eb = surfpoint(x2,y1)
      val ec = surfpoint(x1,y2)
      val ed = surfpoint(x2,y2)
      in triangle 0 (triangle 0 scn ea eb ed) ea ec ed
      end

    fun stripe y1 y2 scn = let
      fun squares 0 scn = scn
        | squares n scn = squares (n-1) (
        print "."; square (itr(n-5)/4.0) y1 (itr(n-4)/4.0) y2 scn
        )
      in squares 8 scn
      end
       
    fun stripes 0 scn = scn
      | stripes n scn = stripes (n-1)(stripe (itr(n-5)/4.0) (itr(n-4)/4.0) scn)

    in stripes 8 nil
    end

  (* geodezyjne *)
  
  (* obliczamy, na jak dlugi odcinek przechodzi odcinek jednostkowy *)
  (* (scale jest podobienswem)                                      *)
  val zoom = let
    val zero = scale(0.0,0.0,0.0)
    val one  = scale(1.0,0.0,0.0)
    val un = sub3(one, zero)
    in R3.norm un
    end
    
  fun geodesic (uv0,duv,length) = let
  
    (* przy kazdym kroku wartosc time zwieksza sie o... *)
    val step = (1.0/zoom/length)

    fun continue (sp0 as (uv0 as (u0,v0),p0)) (duv as (du,dv)) time scn = let
      fun str() = "("^(rts u0)^","^(rts v0)^") ("^(rts du)^","^(rts dv)^")"

      val envnt = setc VT time (envn sp0)
      (* jesli nalezy, to dokoncz z drugiej strony *)
      fun rewrap uv =
         if getc envnt VWrap > 0.0
	 then continue (surfpoint uv) duv time scn 
	 else (print("End: "^(str())^", t="^(rts time)^"\n"); scn)
      (* sprawdz, czy nie nalezy skonczyc albo uzyc rewrap *)
      in if time>1.0 then (print("End: "^str()^"\n"); scn) else 
         if (u0< ~1.0) then rewrap (u0+2.0,v0) else
         if (u0>  1.0) then rewrap (u0-2.0,v0) else
         if (v0< ~1.0) then rewrap (u0,v0+2.0) else
         if (v0>  1.0) then rewrap (u0,v0-2.0) else let

      (* dodaj piksel do sceny *)
      val scn1 = let
        val xx = Real.round (#1(scale(rotate(p0))))
        val yy = Real.round (#2(scale(rotate(p0))))
        val zz = #3(scale(rotate(p0)))
        in if (xx>=0) andalso (xx<CX) andalso (yy>=0) andalso (yy<CY)
	  then addpoint 1 (xx, yy, zz, (getcol envnt VCGeo, 0.0)) scn
	  else scn
        end;

      (* przeskalowujemy duv w ten sposob, zeby odcinek mial na ekranie   *)
      (* dlugosc 1 piksela (na wszelki wypadek robimy to po kazdym kroku) *)
      
      (* rozniczka w uv0 wzdluz duv ... *)
      val df = add3(mul3(du, getU uv0), mul3(dv, getV uv0));
      val dist = R3.norm df;
      val zom = 1.0 / (dist * zoom);
      val duv = mul2(zom, duv);
      
      (* bedzie uv0 w nastepnym kroku *)
      val uv1 = add2(uv0, duv);

      (* bedzie time w nastepnym kroku *)
      val time1 = time + step;
      
      (* surfacepoint w sp1 *)
      val sp1 as (_, p1)= surfpoint uv1;
      
      (* zgodnie z definicja geodezyjnej, powinna ona poruszac sie w ten *)
      (* sposob, ze obrazem "nastepnego" punktu na powierzchni bedzie    *)
      (* p1+(p1-p0) plus jakis wektor ortogonalny do powierzchni w p1.   *)
      (* Znajdujemy taki punkt przez kolejne przyblizenia:               *)
      
      (* dazymy do ... *)
      val p2 = sub3(mul3(2.0, p1), p0);
      
      val M = R3.inverse (getU uv1, getV uv1, getorthogonal sp1)
      (* nastepny krok ... *)
      fun step uvn = let
        val (_,pn) = surfpoint uvn
	val (du,dv,dn) = R3.transform M (sub3(p2, pn))
	in add2(uvn,(du,dv))
	end
      fun iterate f 0 x = x | iterate f n x = iterate f (n-1) (f x)
      (* 10 krokow i gotowe *)
      val uv2 = iterate step 10 (sub2(mul2(2.0, uv1), uv0))
      (* duv1 jako roznica uv2-uv1 *)
      val duv1 = sub2(uv2, uv1)
      
      (* nastepny krok *)
      in continue sp1 duv1 time1 scn1
      end end

    in continue (surfpoint uv0) duv 0.0 nil
    end

  fun point pt = let
    val xx = Real.round (#1(scale(rotate(pt))))
    val yy = Real.round (#2(scale(rotate(pt))))
    val zz = #3(scale(rotate(pt)))
    val env2 = setc VX (#1 pt) (setc VY (#2 pt) (setc VZ (#3 pt) env))
    in if (xx>=0) andalso (xx<CX) andalso (yy>=0) andalso (yy<CY)
      then addpoint 0 (xx, yy, zz, (getcol env2 VCPnt, 0.0)) nil
      else nil
    end;

  fun curve () = let
    type curvepoint = real * R3.Vector.t
    fun tenv t = setc VT t env
    fun cpoint t : curvepoint = let
      fun ev v = getc (tenv t) v
      in (t, (ev VX, ev VY, ev VZ))
      end
    fun vts (x,y,z) = "("^(rts x)^","^(rts y)^","^(rts z)^")"
    fun segment (c0 as (t0,p0)) (c1 as (t1,p1)) scn = let
      val l = R3.norm(sub3(p1,p0))
      in if l < 1.0/zoom
        then let
          val xx = Real.round (#1(scale(rotate(p0))))
          val yy = Real.round (#2(scale(rotate(p0))))
          val zz = #3(scale(rotate(p0)))
          val env2 = setc VX (#1 p0) (setc VY (#2 p0) 
                     (setc VZ (#3 p0) (tenv t0)))
          in if (xx>=0) andalso (xx<CX) andalso (yy>=0) andalso (yy<CY)
            then addpoint 1 (xx, yy, zz, (getcol env2 VCCrv, 0.0)) scn
            else scn
          end
        else let 
          val cm = cpoint ((t0+t1)/2.0)
          in segment c0 cm (segment cm c1 scn)
          end
      end
    fun segments 0 scn = scn
      | segments n scn = segments (n-1) (
        print "."; segment
          (cpoint ((itr(n-17))/32.0))
          (cpoint ((itr(n-16))/32.0)) 
          scn
        )
    in segments 64 nil
    end

  end;
  

