(* Ten modul zawiera definicje zwiazane z przestrzeniami liniowymi. *)

(* po prostu przestrzen liniowa *)
signature VSpace = sig
  eqtype t;
  val zero: t;
  val add: t*t -> t;
  val sub: t*t -> t;
  val mul: real*t -> t;
  
  (* aksjomaty przestrzeni liniowej *)
  axiom Forall (x,y) => sub(x,y) = add(x, mul(~1.0, y));
  axiom Forall (x,y,z) => add(add(x,y),z) = add(x,add(y,z));
  axiom Forall (x,y) => add(x,y) = add(y,x);
  axiom Forall (x,y,a) => mul(a,add(x,y)) = add(mul(a,x), mul(a,y));
  axiom Forall (x,a,b) => add(mul(a,x),mul(b,x)) = mul(a+b,x);
  axiom Forall (x,a,b) => mul(a,mul(b,x)) = mul(a*b, x);
  axiom Forall (x) => mul(1.0,x) = x;
  axiom Forall (x) => mul(0.0,x) = zero;
  end;

(* przestrzen wektorowa wraz z macierzami i podstawowymi operacjami *)
signature VectorSpace = sig

  (* kolumna lub rzad, zlozona niekoniecznie ze skalarow *)
  type 'a up;

  (* jesli potrzebujemy rownosci... *)
  eqtype ''a equp;
  
  sharing type up = equp;

  (* iloczyn tensorowy przez V *)
  functor Up: functor (V: VSpace) -> VSpace;
  
  (* wektory (elementy V) *)
  structure Vector: VSpace where type t = real equp;
  
  (* macierze (jako rzad wektorow) *)
  structure Matrix: VSpace where type t = Vector.t equp;
  
  (* przeksztalcenia liniowe na V *)
  type transform = Vector.t -> Vector.t;
  
  (* przeksztalcenie identycznosciowe *)
  val Id: transform;
  axiom Forall v => Id v = v;
  
  (* zamiana macierzy na przeksztalcenie liniowe i odwrotnie *)
  val transform: Matrix.t -> transform;
  val matrix: transform -> Matrix.t;
  
  axiom Forall M => matrix (transform M) = M;
  axiom Forall L => Forall v => transform (matrix L) v = L v;
  
  (* macierz odwrotna *)
  val inverse: Matrix.t -> Matrix.t;
  axiom Forall M => Forall v => 
      (v = ((transform M) o (transform (inverse M))) v) andalso
      (v = ((transform (inverse M)) o (transform M)) v);

  (* wyznacznik macierzy *)
  val det: Matrix.t -> real;

  axiom det (matrix Id) = 1.0;
  axiom Forall (L1,L2) =>
    det (matrix (L1 o L2)) = det (matrix L1) * det (matrix L2);

  (* slad macierzy *)
  val tr: Matrix.t -> real;

  (* norma wektora *)
  val norm: Vector.t -> real;

  (* iloczyn skalarny *)
  val dotprod: Vector.t -> Vector.t -> real;

  (* normalizacja wektora *)
  val normalize: Vector.t -> Vector.t;
  
  end;

(***************************************************************************)

(* cialo skalarow jako 1-wymiarowa przestrzen liniowa *)
structure Scalar : VSpace = struct
  type t = real;
  val zero = 0.0;
  fun add(x,y) = x+y;
  fun sub(x,y) = x-y;
  fun mul(x,y) = x*y;
  (* ... aksjomaty mozna udowodnic ... *)
  end;

(* dwu- i trojwymiarowe przestrzenie VectorSpace *)

structure R2 : VectorSpace where type 'a up = 'a*'a    = struct

  type 'a up = 'a * 'a;
  eqtype ''a equp = ''a up;

  (* kwadrat kartezjanski przestrzeni *)
  functor Up (V: VSpace) : VSpace = struct
    type t = V.t up;
    val zero = (V.zero, V.zero);
    fun add ((vx,vy),(wx,wy)) = (V.add(vx,wx), V.add(vy,wy));
    fun sub ((vx,vy),(wx,wy)) = (V.sub(vx,wx), V.sub(vy,wy));
    fun mul (n,(x,y)) = (V.mul(n,x), V.mul(n,y));
    (* jesli V spelnia aksjomaty, to Up(V) tez *)
    end;

  structure Vector = Up(Scalar);
  
  structure Matrix = Up(Vector);
  
  type transform = Vector.t -> Vector.t;
  
  fun Id x = x;
  
  fun transform ((a,b),(c,d)) (x,y:real) = (a*x+c*y, b*x+d*y);

  fun matrix f = (f(1.0,0.0), f(0.0,1.0));
  
  fun det ((a,b),(c,d)) : real = a*d-b*c;

  fun inverse (M as ((a,b),(c,d))) = Matrix.mul((1.0/det M), ((d,~b),(~c,a)));

  fun tr ((a,b),(c,d)) : real = a+d;
    
  fun norm (x,y) = Math.sqrt(x*x+y*y);

  fun dotprod (x1,y1) (x2,y2):real = x1*x2+y1*y2;
  
  fun normalize (v as (vx,vy)) = let
    val n = norm v
    in if n=0.0 then v else (vx/n, vy/n) end;
      
  end;
  
structure R3 : VectorSpace
  where type 'a up = 'a*'a*'a = struct

  type 'a up = 'a * 'a * 'a;
  type ''a equp = ''a up;

  (* szescian kartezjanski przestrzeni *)
  functor Up (V: VSpace) : VSpace = struct
    type t = V.t up;
    val zero = (V.zero, V.zero, V.zero);
    fun add ((vx,vy,vz),(wx,wy,wz)) = (V.add(vx,wx), V.add(vy,wy), V.add(vz,wz));
    fun sub ((vx,vy,vz),(wx,wy,wz)) = (V.sub(vx,wx), V.sub(vy,wy), V.sub(vz,wz));
    fun mul (n,(x,y,z)) = (V.mul(n,x), V.mul(n,y), V.mul(n,z));
    end;

  type scalar = Scalar.t;
  
  structure Vector = Up(Scalar);
  structure Matrix = Up(Vector);
  
  type transform = Vector.t -> Vector.t;
  
  fun Id x = x;
  
  fun transform ((a,b,c),(d,e,f),(g,h,i)) (x,y,z:real) =
    (a*x+d*y+g*z, b*x+e*y+h*z, c*x+f*y+i*z);

  fun matrix f = (f(1.0,0.0,0.0), f(0.0,1.0,0.0), f(0.0,0.0,1.0));
  
  fun det ((a,b,c),(d,e,f),(g,h,i)) : real = 
    a*e*i-a*f*h-b*d*i+b*f*g+c*d*h-c*e*g;

  fun inverse (M as ((a,b,c),(d,e,f),(g,h,i))) = Matrix.mul((1.0/det M),
    ((e*i-f*h,h*c-i*b,b*f-c*e),
     (f*g-d*i,i*a-g*c,c*d-a*f),
     (d*h-e*g,g*b-h*a,a*e-b*d)
     ));

  fun tr ((a,b,c),(d,e,f),(g,h,i)) : real = a+e+i;
    
  fun norm (x,y,z) = Math.sqrt(x*x+y*y+z*z);

  fun dotprod (x1,y1,z1) (x2,y2,z2):real = x1*x2+y1*y2+z1*z2;
  
  fun normalize (v as (vx,vy,vz)) = let
    val n = norm v
    in if n=0.0 then (vx,vy,vz) else (vx/n, vy/n, vz/n)
    end;

  end;

(* iloczyn skalarny dwoch wektorow trojwymiarowych *)  
fun crossprod (x1,y1,z1) (x2,y2,z2) : R3.Vector.t = 
  (y1*z2-z1*y2, z1*x2-x1*z2, x1*y2-y1*x2);

(* kolory to tez przestrzen wektorowa R3 *)

structure Color = R3;

(* uzyteczne skroty *)

val mul3 = R3.Vector.mul;
val add3 = R3.Vector.add;
val sub3 = R3.Vector.sub;

val mul2 = R2.Vector.mul;
val add2 = R2.Vector.add;
val sub2 = R2.Vector.sub;
